from ..person import Person
from django import forms
from django.forms import DateInput
from rest_framework import serializers
from ...patient_examination import PatientExamination
from ...data_file import ReportFile
from django.db import models
from faker import Faker
import random
from datetime import datetime

class Patient(Person):
    """
    A class representing a patient.

    Attributes inhereted from Person:
        first_name (str): The first name of the patient.
        last_name (str): The last name of the patient.
        dob (datetime.date): The date of birth of the patient.
        gender (Foreign Key): The gender of the patient.
        email (str): The email address of the patient.
        phone (str): The phone number of the patient.

    """
    center = models.ForeignKey("Center", on_delete=models.CASCADE, blank=True, null=True)
    

    def __str__(self):
        return self.first_name + " " + self.last_name + " (" + str(self.dob) + ")"
    
    def get_unmatched_report_files(self): #field: self.report_files; filter: report_file.patient_examination = None
        '''Returns all report files for this patient that are not matched to a patient examination.'''

        return self.reportfile_set.filter(patient_examination=None)

    def get_unmatched_video_files(self): #field: self.videos; filter: video.patient_examination = None
        '''Returns all video files for this patient that are not matched to a patient examination.'''
        return self.videos.filter(patient_examination=None)

    def get_patient_examinations(self): #field: self.patient_examinations
        '''Returns all patient examinations for this patient ordered by date (most recent is first).'''
        return self.patient_examinations.order_by('-date')
    
    def create_examination_by_report_file(self, report_file:ReportFile):
        '''Creates a patient examination for this patient based on the given report file.'''
        patient_examination = PatientExamination(patient=self, report_file=report_file)
        patient_examination.save()
        return patient_examination
    
    @classmethod
    def get_random_gender(self, p_male=0.5, p_female=0.5):
        """
        Get a Gender object by name (male, female) from the database with given probability.

        :param p_male: Probability of selecting 'male' gender.
        :param p_female: Probability of selecting 'female' gender.
        :return: Gender object selected based on given probabilities.
        """
        from endoreg_db.models import Gender
        
        # Extract names and probabilities
        gender_names = ["male", "female"]
        probabilities = [0.5, 0.5]
        
        # Debug: print the names and probabilities
        print(f"Gender names: {gender_names}")
        print(f"Probabilities: {probabilities}")
        
        # Select a gender based on the given probabilities
        selected_gender = random.choices(gender_names, probabilities)[0]
        # Debug: print the selected gender
        print(f"Selected gender: {selected_gender}")
        
        # Fetch the corresponding Gender object from the database
        gender_obj = Gender.objects.get(name=selected_gender)
        
        return gender_obj


    @classmethod
    def get_random_age(self, 
            min_age = 55,
            max_age = 90,
            mean_age = 65,
            std_age = 10,
            distribution = "normal"
    ):
        """
        Get a random age based on the given distribution.

        :param min_age: Minimum age.
        :param max_age: Maximum age.
        :param mean_age: Mean age.
        :param std_age: Standard deviation of the age.
        :param distribution: Distribution of the age.
        :return: Random age based on the given distribution.
        """
        if distribution == "normal":
            age = int(random.normalvariate(mean_age, std_age))
        else:
            age = int(random.uniform(min_age, max_age))
        
        return age
    
    @classmethod
    def get_dob_from_age(self, age, current_date=None):
        """
        Get a date of birth based on the given age and current date.

        :param age: Age of the patient.
        :param current_date: Current date.
        :return: Date of birth based on the given age and current date.
        """
        if current_date is None:
            current_date = datetime.now()
        dob = current_date.replace(year=current_date.year - age).date()

        # TODO
        # randomize the day and month by adding a random number of days (0-364) to the date

        return dob
        
    @classmethod
    def get_random_name_for_gender(self, gender_obj, locale="de_DE"):
        gender = gender_obj.name
        fake = Faker(locale)
        
        if gender == "male":
            first_name = fake.first_name_male()
            last_name = fake.last_name_male()

        else: 
            first_name = fake.first_name_female()
            last_name = fake.last_name_female()

        return last_name, first_name
    
    @classmethod
    def create_generic(self, center="gplay_case_generator"):
        """
        Create a generic patient with random attributes.

        :param center: The center of the patient.
        :return: The created patient.
        """
        from endoreg_db.models import Center
        gender = Patient.get_random_gender()
        last_name, first_name = Patient.get_random_name_for_gender(gender)

        age = Patient.get_random_age()
        dob = Patient.get_dob_from_age(age)

        center = Center.objects.get(name=center)

        patient = Patient.objects.create(
            first_name=first_name,
            last_name=last_name,
            dob=dob
        )
        patient.save()
        return patient

    def age(self):
        """
        Get the age of the patient.

        :return: The age of the patient.
        """
        # calculate correct age based on current date including day and month
        current_date = datetime.now()
        dob = self.dob
        age = current_date.year - dob.year - ((current_date.month, current_date.day) < (dob.month, dob.day))
        return age

    def create_lab_sample(self, sample_type="generic", date=None, save=True):
        """
        Create a lab sample for this patient.

        :param sample_type: The sample type. Should be either string of the sample types
            name or the sample type object. If not set, the default sample type ("generic") is used.
        :param date: The date of the lab sample.
        :return: The created lab sample.
        """
        from endoreg_db.models import PatientLabSample, PatientLabSampleType
        if date is None:
            date = datetime.now()

        if isinstance(sample_type, str):
            sample_type = PatientLabSampleType.objects.get(name=sample_type)
            assert sample_type is not None, f"Sample type with name '{sample_type}' not found."#
        elif not isinstance(sample_type, PatientLabSampleType):
            raise ValueError("Sample type must be either a string or a PatientLabSampleType object.")
        
        patient_lab_sample = PatientLabSample.objects.create(
            patient=self,
            sample_type=sample_type,
            date=date
        )

        if save:
            patient_lab_sample.save()

        return patient_lab_sample

class PatientForm(forms.ModelForm):
    class Meta:
        model = Patient
        fields = '__all__'
        widgets = {
            'dob': DateInput(attrs={'type': 'date'}),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.values():
            field.widget.attrs['class'] = 'form-control'



from django.db import models
from ..label.label import Label
from ..annotation.image_classification import ImageClassificationAnnotation

class LegacyImage(models.Model):
    # Add any other fields you need to store frame-related information
    image = models.ImageField(upload_to="legacy_images")  # Or some other field type, depending on how you're storing the frame
    suffix = models.CharField(max_length=255)
    # ImageClassificationAnnotation has a foreign key to this model (related name: image_classification_annotations)

    def get_classification_annotations(self):
        """
        Get all image classification annotations for this image.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_image=self)
    
    def get_classification_annotations_by_label(self, label:Label):
        """
        Get all image classification annotations for this image with the given label.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_image=self, label=label)
    
    def get_classification_annotations_by_value(self, value:bool):
        """
        Get all image classification annotations for this image with the given value.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_image=self, value=value)
    
    def get_classification_annotations_by_label_and_value(self, label:Label, value:bool):
        """
        Get all image classification annotations for this image with the given label and value.
        """
        return ImageClassificationAnnotation.objects.filter(legacy_image=self, label=label, value=value)
    



from django.db import models

class ModelMetaManager(models.Manager):
    # natural key is name and version
    def get_by_natural_key(self, name, version):
        return self.get(name=name, version=version)

class ModelMeta(models.Model):
    name = models.CharField(max_length=255)
    version = models.CharField(max_length=255)
    type = models.ForeignKey("ModelType", on_delete=models.CASCADE, related_name="models")
    labelset = models.ForeignKey("LabelSet", on_delete=models.CASCADE, related_name="models")
    weights = models.FileField(upload_to='weights/')
    
    description = models.TextField(blank=True, null=True)
    date_created = models.DateTimeField(auto_now_add=True)
    objects = ModelMetaManager()

    def natural_key(self):
        return (self.name, self.version)

    def __str__(self):
        return f"{self.name} (v: {self.version})"




from rest_framework import serializers
from endoreg_db.models import LegacyBinaryClassificationAnnotationTask, BinaryClassificationAnnotationTask, ImageClassificationAnnotation

class LegacyBinaryClassificationAnnotationTaskSerializer(serializers.ModelSerializer):
    class Meta:
        model = LegacyBinaryClassificationAnnotationTask
        fields = '__all__'

class BinaryClassificationAnnotationTaskSerializer(serializers.ModelSerializer):
    class Meta:
        model = BinaryClassificationAnnotationTask
        fields = '__all__'

class ImageClassificationAnnotationSerializer(serializers.ModelSerializer):
    class Meta:
        model = ImageClassificationAnnotation
        fields = '__all__'



from django.db import models
from django.utils.translation import gettext_lazy as _

class PatientMedication(models.Model):
    patient = models.ForeignKey("Patient", on_delete= models.CASCADE)
    medication_indication = models.ForeignKey(
        "MedicationIndication", on_delete=models.CASCADE,
        related_name="patient_medications", null=True
    )

    medication_schedules = models.ManyToManyField(
        'MedicationSchedule'
    )
    unit = models.ForeignKey('Unit', on_delete=models.CASCADE)
    dosage = models.JSONField()
    active = models.BooleanField(default=True)

    objects = models.Manager()

    class Meta:
        verbose_name = _('Patient Medication')
        verbose_name_plural = _('Patient Medications')

    @classmethod
    def create_by_patient_and_indication(cls, patient, medication_indication):
        from endoreg_db.models import MedicationIndication
        medication_indication: MedicationIndication
        patient_medication = cls.objects.create(patient=patient, medication_indication=medication_indication)
        patient_medication.save()
        patient_medication.set_schedules_from_indication()
        
        return patient_medication

    def set_schedules_from_indication(self):
        schedules = self.medication_indication.medication_schedules.all()
        self.medication_schedules.set(schedules)
        self.save()

    def __str__(self):
        indication = self.medication_indication
        schedules = self.medication_schedules.all()
        return f'{indication} - {schedules}'
    
    


from itertools import product
from django.db import models

class CenterProduct(models.Model):
    product = models.ForeignKey("Product", on_delete=models.CASCADE)
    date_used = models.DateField()
    center = models.ForeignKey("Center", on_delete=models.CASCADE)


    def get_product_group(self):
        product = self.product
        product_group = product.product_group

        return product_group
    
    def get_reference_product(self):
        product_group = self.get_product_group()
        reference_product = product_group.reference_product

        return reference_product
    
    def get_product_weight(self):
        product = self.product
        weight, unit = product.get_product_weight()

        return weight, unit
    
    def get_package_weight(self):
        product = self.product
        weight, unit = product.get_package_weight()

        return weight, unit
    
    


from rest_framework import serializers
from endoreg_db.models import (
    Label,
    LabelSet,
    LabelType,
)

class LabelTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = LabelType
        fields = '__all__'

class LabelSetSerializer(serializers.ModelSerializer):
    class Meta:
        model = LabelSet
        fields = '__all__'

class LabelSerializer(serializers.ModelSerializer):
    class Meta:
        model = Label
        fields = '__all__'
        



from rest_framework import serializers
from endoreg_db.models import (
    ImageClassificationPrediction,
    LegacyVideoPredictionMeta
)

class ImageClassificationPredictionSerializer(serializers.ModelSerializer):
    class Meta:
        model = ImageClassificationPrediction
        fields = '__all__'

class LegacyVideoPredictionMetaSerializer(serializers.ModelSerializer):
    class Meta:
        model = LegacyVideoPredictionMeta
        fields = '__all__'



# models/data_file/import_classes/raw_pdf.py
# django db model "RawPdf"
# Class to store raw pdf file using django file field
# Class contains classmethod to create object from pdf file
# objects contains methods to extract text, extract metadata from text and anonymize text from pdf file uzing agl_report_reader.ReportReader class
# ------------------------------------------------------------------------------

from django.db import models
from django.core.files.storage import FileSystemStorage
from django.core.files import File
from django.conf import settings
from django.utils import timezone
from django.core.exceptions import ValidationError
from django.core.validators import FileExtensionValidator
from endoreg_db.utils.file_operations import get_uuid_filename

from agl_report_reader.report_reader import ReportReader

from endoreg_db.utils.hashs import get_pdf_hash
from ..metadata import SensitiveMeta

# setup logging to pdf_import.log
import logging
logger = logging.getLogger('pdf_import')

import shutil

# get pdf location from settings, default to ~/erc_data/raw_pdf and create if not exists
PSEUDO_DIR_RAW_PDF = getattr(settings, 'PSEUDO_DIR_RAW_PDF', settings.BASE_DIR / 'erc_data/raw_pdf')

class RawPdfFile(models.Model):
    file = models.FileField(
        upload_to='raw_pdf/',
        validators=[FileExtensionValidator(allowed_extensions=['pdf'])],
        storage=FileSystemStorage(location=PSEUDO_DIR_RAW_PDF.resolve().as_posix()),
    )

    pdf_hash = models.CharField(max_length=255, unique=True)
    pdf_type = models.ForeignKey('PdfType', on_delete=models.CASCADE)
    center = models.ForeignKey('Center', on_delete=models.CASCADE)

    state_report_processing_required = models.BooleanField(default = True)
    state_report_processed = models.BooleanField(default=False)

    # report_file = models.OneToOneField("ReportFile", on_delete=models.CASCADE, null=True, blank=True)
    sensitive_meta = models.OneToOneField(
        'SensitiveMeta',
        on_delete=models.CASCADE,
        related_name='raw_pdf_file',
        null=True,
        blank=True,
    )

    text = models.TextField(blank=True, null=True)
    anonymized_text = models.TextField(blank=True, null=True)

    raw_meta = models.JSONField(blank=True, null=True)

    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        str_repr = f"RawPdfFile: {self.file.name}"
        return str_repr

    @classmethod
    def create_from_file(
        cls,
        file_path,
        center_name,
        pdf_type_name,
        destination_dir,
        save=True,
    ):
        from endoreg_db.models import PdfType, Center
        logger.info(f"Creating RawPdfFile object from file: {file_path}")
        original_file_name = file_path.name

        new_file_name, uuid = get_uuid_filename(file_path)

        if not destination_dir.exists():
            destination_dir.mkdir(parents=True)

        pdf_hash = get_pdf_hash(file_path)

        # check if pdf file already exists
        if cls.objects.filter(pdf_hash=pdf_hash).exists():
            logger.warning(f"RawPdfFile with hash {pdf_hash} already exists")
            return None
        
        assert pdf_type_name is not None, "pdf_type_name is required"
        assert center_name is not None, "center_name is required"

        pdf_type = PdfType.objects.get(name=pdf_type_name)
        center = Center.objects.get(name=center_name)

        new_file_path = destination_dir / new_file_name

        logger.info(f"Copying file to {new_file_path}")
        success = shutil.copy(file_path, new_file_path)
         
        # validate copy operation by comparing hashs
        assert get_pdf_hash(new_file_path) == pdf_hash, "Copy operation failed"

        raw_pdf = cls(
            file=new_file_path.resolve().as_posix(),
            pdf_hash=pdf_hash,
            pdf_type=pdf_type,
            center=center,
        )
        logger.info(f"RawPdfFile object created: {raw_pdf}")

        # remove source file
        file_path.unlink()
        logger.info(f"Source file removed: {file_path}")

        if save:
            raw_pdf.save()
        

        return raw_pdf

    def process_file(self, verbose = False):
        
        pdf_path = self.file.path
        rr_config = self.get_report_reader_config()

        rr = ReportReader(**rr_config) #FIXME In future we need to pass a configuration file 
        # This configuration file should be associated with pdf type 

        text, anonymized_text, report_meta = rr.process_report(pdf_path, verbose=verbose)
        if not self.sensitive_meta:
            sensitive_meta = SensitiveMeta.create_from_dict(report_meta)
            sensitive_meta.save()
            self.sensitive_meta = sensitive_meta

        else: 
            # update existing sensitive meta
            sensitive_meta = self.sensitive_meta
            sensitive_meta.update_from_dict(report_meta)

        return text, anonymized_text, report_meta
    
    def update(self, save=True, verbose = True):
        try:
            self.text, self.anonymized_text, self.raw_meta = self.process_file(verbose = verbose)
            self.state_report_processed = True
            self.state_report_processing_required = False
        
            if save: 
    
                self.save()

            return True

        except:
            logger.error(f"Error processing file: {self.file.path}")
            return False

    def save(self, *args, **kwargs):
        if not self.file.name.endswith('.pdf'):
            raise ValidationError('Only PDF files are allowed')
        
        if not self.pdf_hash:
            self.pdf_hash = get_pdf_hash(self.file.path)

        super().save(*args, **kwargs)


    def get_report_reader_config(self):
        if self.pdf_type.endoscope_info_line:
            endoscope_info_line = self.pdf_type.endoscope_info_line.value
        else:
            endoscope_info_line = None
        settings_dict = {
            "locale": "de_DE",
            "employee_first_names": [_.name for _ in self.center.first_names.all()],
            "employee_last_names": [_.name for _ in self.center.last_names.all()],
            "text_date_format":'%d.%m.%Y',
            "flags": {
                "patient_info_line": self.pdf_type.patient_info_line.value,
                "endoscope_info_line": endoscope_info_line,
                "examiner_info_line": self.pdf_type.examiner_info_line.value,
                "cut_off_below": [_.value for _ in self.pdf_type.cut_off_below_lines.all()],
                "cut_off_above": [_.value for _ in self.pdf_type.cut_off_above_lines.all()],
            }
        }

        return settings_dict



from .unit import Unit
from .information_source import InformationSource
from .center import *
from .report_reader import ReportReaderConfig, ReportReaderFlag
from .emission import *
from .persons import *
from .network import *
from .logging import *

from .case_template import *
# from .other.distribution import (
#     SingleCategoricalValueDistribution,
#     NumericValueDistribution,
#     MultipleCategoricalValueDistribution,
#     DateValueDistribution,
# )
from .rules import (
    Rule,
    RuleType,
    Ruleset,
    RuleApplicator,
    Rule
)

from .disease import Disease, DiseaseClassification, DiseaseClassificationChoice
from .event import Event
from .laboratory import LabValue
from .medication import *

from .examination import (
    Examination,
    ExaminationType,
    ExaminationTime,
    ExaminationTimeType,
)
from .data_file import *

from .patient_examination import PatientExamination

from .label import Label, LabelType, LabelSet

from .annotation import (
    ImageClassificationAnnotation,
    LegacyBinaryClassificationAnnotationTask,
    BinaryClassificationAnnotationTask,
)

from .legacy_data import (
    LegacyImage,
)

from .other import *

from .product import *

from .ai_model import (
    ModelMeta,
    ModelType,
    ActiveModel,
)

from .quiz import *

from .prediction import (
    ImageClassificationPrediction,
    LegacyVideoPredictionMeta,
    VideoPredictionMeta,
)

from .hardware import (
    EndoscopyProcessor,
    Endoscope,
    EndoscopeType,
)

from .questionnaires import TtoQuestionnaire



from django.core.management import BaseCommand, call_command

class Command(BaseCommand):
    help = 'Run all data loading commands in the correct order'

    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='Display verbose output for all commands',
        )

    def handle(self, *args, **options):
        # verbose = options['verbose']
        verbose = True

        self.stdout.write(self.style.SUCCESS("Populating base db models with data..."))

        # Run the load_network_data command
        self.stdout.write(self.style.SUCCESS("Running load_network_data..."))
        call_command('load_network_data', verbose=verbose)

        # Run the load_logging_data command
        self.stdout.write(self.style.SUCCESS("Running load_logging_data..."))
        call_command('load_logging_data', verbose=verbose)

        # Run the load_profession_data command
        self.stdout.write(self.style.SUCCESS("Running load_profession_data..."))
        call_command('load_profession_data', verbose=verbose)

        # run the load_gender_data command
        self.stdout.write(self.style.SUCCESS("Running load_gender_data..."))
        call_command('load_gender_data', verbose=verbose)

        # Run the load_disease data command
        self.stdout.write(self.style.SUCCESS("Running load_disease_data..."))
        call_command('load_disease_data', verbose=verbose)

        # Run the load_disease_classification data command
        self.stdout.write(self.style.SUCCESS("Running load_disease_classification_data..."))
        call_command('load_disease_classification_data', verbose=verbose)

        # Run the load_disease_classification_choices data command
        self.stdout.write(self.style.SUCCESS("Running load_disease_classification_choices_data..."))
        call_command('load_disease_classification_choices_data', verbose=verbose)

        # rund the load_user_groups command
        self.stdout.write(self.style.SUCCESS("Running load_user_groups..."))
        call_command('load_user_groups', verbose=verbose)

        # run the load_report_reader_flag command
        self.stdout.write(self.style.SUCCESS("Running load_report_reader_flag..."))
        call_command('load_report_reader_flag', verbose=verbose)

        # Run the load_pdf_type_data command
        self.stdout.write(self.style.SUCCESS("Running load_pdf_type_data..."))
        call_command('load_pdf_type_data', verbose=verbose)

        # Run the load_endoscopy_processor_data command
        self.stdout.write(self.style.SUCCESS("Running load_endoscopy_processor_data..."))
        call_command('load_endoscopy_processor_data', verbose=verbose)

        # Run the load_endoscope_type_data command
        self.stdout.write(self.style.SUCCESS("Running load_endoscope_type_data..."))
        call_command('load_endoscope_type_data', verbose=verbose)

        # Run the load_unit_data command
        self.stdout.write(self.style.SUCCESS("Running load_unit_data..."))
        call_command('load_unit_data', verbose=verbose)

        # Run the load_lab_value_data command
        self.stdout.write(self.style.SUCCESS("Running load_lab_value_data..."))
        call_command("load_lab_value_data", verbose=verbose)

        # Run the load_information_source command
        self.stdout.write(self.style.SUCCESS("Running load_information_source..."))
        call_command('load_information_source', verbose=verbose)

        # Run the load_center_data command
        self.stdout.write(self.style.SUCCESS("Running load_center_data..."))
        call_command('load_center_data', verbose=verbose)

        # Run the load_examination_data command
        self.stdout.write(self.style.SUCCESS("Running load_examination_data..."))
        call_command('load_examination_data', verbose=verbose)

        # Run the load_label_data command
        self.stdout.write(self.style.SUCCESS("Running load_label_data..."))
        call_command('load_label_data', verbose=verbose)

        # Run the load_ai_model_data command
        self.stdout.write(self.style.SUCCESS("Running load_ai_model_data..."))
        call_command('load_ai_model_data', verbose=verbose)

        # Run the load_event_data command
        self.stdout.write(self.style.SUCCESS("Running load_event_data..."))
        call_command('load_event_data', verbose=verbose)

        # run the load_medication_data command
        self.stdout.write(self.style.SUCCESS("Running load_medication_data..."))
        call_command('load_medication_data', verbose=verbose)

        # Run the load_medication_indication_type_data command
        self.stdout.write(self.style.SUCCESS("Running load_medication_indication_type_data..."))
        call_command('load_medication_indication_type_data', verbose=verbose)

        # Run the load_medication_intake_time_data command
        self.stdout.write(self.style.SUCCESS("Running load_medication_intake_time_data..."))
        call_command('load_medication_intake_time_data', verbose=verbose)

        # Run the load_medication_schedule_data command
        self.stdout.write(self.style.SUCCESS("Running load_medication_schedule_data..."))
        call_command('load_medication_schedule_data', verbose=verbose)

        # Run the load_medication_indication_data command
        self.stdout.write(self.style.SUCCESS("Running load_medication_indication_data..."))
        call_command('load_medication_indication_data', verbose=verbose)

        # Run the load_green_endoscopy_wuerzburg_data command
        self.stdout.write(self.style.SUCCESS("Running load_green_endoscopy_wuerzburg_data..."))
        call_command('load_green_endoscopy_wuerzburg_data', verbose=verbose)

        # Load G-Play Data
        # run the load_distribution_data command
        self.stdout.write(self.style.SUCCESS("Running load_distribution_data..."))
        call_command('load_distribution_data', verbose=verbose)
        # Run the load_g_play_data command
        self.stdout.write(self.style.SUCCESS("Running load_g_play_data..."))
        call_command('load_g_play_data', verbose=verbose)

        # Necessary? Migrate? FIXME
        # # Run the load_active_model_data command
        # self.stdout.write(self.style.SUCCESS("Running load_active_model_data..."))
        # call_command('load_active_model_data', verbose=verbose)
        
        self.stdout.write(self.style.SUCCESS("All data loading commands executed successfully."))



from django.db import models
from .patient import Patient

class PatientEvent(models.Model):
    """
    A class representing an event for a patient.

    Attributes:
        patient (Patient): The patient associated with this event.
        date (datetime.date): The date of the event.
        description (str): A description of the event.
    """
    patient = models.ForeignKey(Patient, on_delete=models.CASCADE)
    event = models.ForeignKey("Event", on_delete=models.CASCADE)
    date_start = models.DateField()
    date_end = models.DateField(blank=True, null=True)
    description = models.TextField(blank=True, null=True)

    last_update = models.DateTimeField(auto_now=True)
    
    def __str__(self):
        return str(self.date_start) + ": " + self.event



from django.db import models

class CaseTemplateRuleTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class CaseTemplateRuleType(models.Model):
    """
    A class representing a case template rule type.

    Attributes:
        name (str): The name of the case template rule type.
        description (str): A description of the case template rule type.

    """
    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, null=True)
    name_en = models.CharField(max_length=255, null=True)
    description = models.TextField(blank=True, null=True)

    objects = CaseTemplateRuleTypeManager()

    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name

class CaseTemplateRuleManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
    def get_default_create_patient_rule(self):
        """Get the default create patient rule."""
        return self.get(name="create-patient-default")
    
    def get_default_create_patient_lab_sample_rules(self):
        """Get the default create patient lab sample rules."""
        return self.get(name="create-patient_lab_sample-default")
    
    def get_default_create_patient_medication_schedule_rule(self):
        """Get the create patient medication schedule rule."""
        r = self.get(name="create-patient_medication_schedule")
        assert r, "No create patient medication schedule rule found"
        return r
    
class CaseTemplateRule(models.Model):
    """
    A class representing a case template rule.

    Attributes:
        name (str): The name of the case template rule.
        case_template (CaseTemplate): The case template to which the rule belongs.
        description (str): A description of the case template rule.
        name_de (str): The name of the case template rule in German.
        name_en (str): The name of the case template rule in English.
        rule_type (CaseTemplateRuleType): The type of the rule.
        parent_model (str): The model on which the rule is applied.
        target_field (str): The field of the parent model on which the rule is applied.
        target_model (str): The model to which the foreign key points (used for foreign key and many-to-many rules).
        rule_values (list): A list of string values for the rule.
        extra_parameters (dict): A dictionary of extra parameters for the rule.
        value_type (CaseTemplateRuleValueType): The type of the rule value.
        chained_rules (QuerySet): The chained rules associated with the current rule.
        single_categorical_value_distribution (SingleCategoricalValueDistribution): The single categorical value distribution for the rule.
        numerical_value_distribution (NumericValueDistribution): The numerical value distribution for the rule.
        multiple_categorical_value_distribution (MultipleCategoricalValueDistribution): The multiple categorical value distribution for the rule.
        date_value_distribution (DateValueDistribution): The date value distribution for the rule.
        objects (CaseTemplateRuleManager): The manager for the CaseTemplateRule model.

    Methods:
        natural_key(): Returns the natural key of the rule.
        __str__(): Returns a string representation of the rule.
        get_rule_type(): Returns the name of the rule type.
        get_chained_rules(): Returns all the chained rules associated with the current rule.
        get_all_downward_chained_rules(): Returns all the chained rules of the current rule, including indirectly chained rules.
        chained_rules_has_self_reference(): Checks if any directly or indirectly chained rules reference this rule, creating a non-terminating loop.
        save(*args, **kwargs): Customizes the save method to check for self-references.

    """
    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, null=True)
    name_en = models.CharField(max_length=255, null=True)
    description = models.TextField(blank=True, null=True)
    rule_type = models.ForeignKey(
        "CaseTemplateRuleType", on_delete=models.CASCADE
    )
    parent_model = models.CharField(max_length=255, blank=True, null=True)
    parent_field = models.CharField(max_length=255, blank=True, null=True)
    target_field = models.CharField(max_length=255, blank=True, null=True)
    target_model = models.CharField(max_length=255, blank=True, null=True)
    rule_values = models.JSONField(blank=True, null=True)
    extra_parameters = models.JSONField(blank=True, null=True)
    value_type = models.ForeignKey("CaseTemplateRuleValueType", on_delete=models.SET_NULL, null=True)
    chained_rules = models.ManyToManyField(
        "CaseTemplateRule",
        related_name="calling_rules"
    )
    single_categorical_value_distribution = models.ForeignKey(
        "SingleCategoricalValueDistribution", on_delete=models.SET_NULL, null=True
    )
    numerical_value_distribution = models.ForeignKey(
        "NumericValueDistribution", on_delete=models.SET_NULL, null=True
    )
    multiple_categorical_value_distribution = models.ForeignKey(
        "MultipleCategoricalValueDistribution", on_delete=models.SET_NULL, null=True
    )
    date_value_distribution = models.ForeignKey(
        "DateValueDistribution", on_delete=models.SET_NULL, null=True
    )
    objects = CaseTemplateRuleManager()

    def natural_key(self):
        """
        Returns the natural key of the rule.

        :return: A tuple representing the natural key of the rule.
        """
        return (self.name,)

    def __str__(self):
        """
        Returns a string representation of the rule.

        :return: A string representation of the rule.
        """
        return self.name

    def get_rule_type(self):
        """
        Returns the name of the rule type.

        :return: The name of the rule type.
        """
        return self.rule_type.name

    def get_chained_rules(self):
        """
        Returns all the chained rules associated with the current rule.

        :return: A queryset of all the chained rules associated with the current rule.
        """
        return self.chained_rules.all()
    
    def get_distribution(self):
        """
        Returns the value distribution of the rule.

        :return: The value distribution of the rule.
        """
        DEBUG = True
        if DEBUG:
            print("Rule: ", self.name)
            print(f"single_categorical_value_distribution: {self.single_categorical_value_distribution}")
            print(f"numerical_value_distribution: {self.numerical_value_distribution}")
            print(f"multiple_categorical_value_distribution: {self.multiple_categorical_value_distribution}")
            print(f"date_value_distribution: {self.date_value_distribution}")


        if self.single_categorical_value_distribution:
            return self.single_categorical_value_distribution
        elif self.numerical_value_distribution:
            return self.numerical_value_distribution
        elif self.multiple_categorical_value_distribution:
            return self.multiple_categorical_value_distribution
        elif self.date_value_distribution:
            return self.date_value_distribution
        return None

    def get_all_downward_chained_rules(self):
        """
        Get all chained rules of the current rule, including indirectly chained rules.

        :return: A set of all chained rules of the current rule.
        """
        all_chained_rules = set()

        def traverse_chained_rules(rule):
            """
            Helper function to recursively traverse chained rules.

            :param rule: The current CaseTemplateRule being checked.
            :return: None
            """
            for chained_rule in rule.chained_rules.all():
                if chained_rule not in all_chained_rules:
                    all_chained_rules.add(chained_rule)
                    traverse_chained_rules(chained_rule)

        # Initialize the traversal starting with the current rule
        traverse_chained_rules(self)
        return all_chained_rules

    def chained_rules_has_self_reference(self):
        """
        Check if any directly or indirectly chained rules reference this rule, creating a non-terminating loop.
        Return a list containing a tuple (first_rule, self_referencing_rule) for each self-reference.
        first_rule is the most upward rule in the chain that references the rule.
        self_referencing_rule is the rule that references the current rule itself.

        :return: A list of tuples representing the self-referencing rules.
        """
        result_list = []

        def traverse_chained_rules(rule, visited_rules):
            """
            Helper function to recursively traverse chained rules and check for self-references.

            :param rule: The current CaseTemplateRule being checked.
            :param visited_rules: A list of tuples representing the path of rules visited so far.
            :return: None
            """
            for chained_rule in rule.chained_rules.all():
                if chained_rule == self:
                    # A self-reference is detected
                    first_rule = visited_rules[0][0]
                    self_referencing_rule = visited_rules[-1][1]
                    result_list.append((first_rule, self_referencing_rule))
                elif chained_rule not in [r[1] for r in visited_rules]:
                    # Continue to check chained rules of the current chained rule
                    traverse_chained_rules(chained_rule, visited_rules + [(rule, chained_rule)])

        # Initialize the traversal starting with the current rule
        traverse_chained_rules(self, [(self, self)])
        return result_list
    
    def get_target_model(self):
        """
        Returns the target model of the rule.

        :return: The target model of the rule.
        """
        from django.apps import apps

        if not self.target_model:
            return None

        try:
            target_model = apps.get_model("endoreg_db", self.target_model)
        except LookupError:
            raise ValueError(f"Model {self.target_model} not found.")
        return target_model
    
    def get_target_field(self):
        """
        Returns the target field of the rule.

        :return: The target field of the rule.
        """
        return self.target_field
    
    def get_parent_model(self):
        """
        Returns the parent model of the rule.

        :return: The parent model of the rule.
        """
        from django.apps import apps

        if not self.parent_model:
            return None

        try:
            self.parent_model = apps.get_model("endoreg_db", self.parent_model)
        except LookupError:
            raise ValueError(f"Model {self.parent_model} not found.")
        return self.parent_model

    # customize the save method to check for self-references
    # def save(self, *args, **kwargs):
    #     # Check for self-references
    #     self_referencing_rules = self.chained_rules_has_self_reference()
    #     if self_referencing_rules:
    #         raise ValueError(f"Self-references detected: {self_referencing_rules}")
    #     super().save(*args, **kwargs)




from endoreg_db.models import ModelMeta, ModelType, ActiveModel
from rest_framework import serializers

class ModelMetaSerializer(serializers.ModelSerializer):
    weights = serializers.FileField(use_url=True)  # use_url=True will provide the file's URI

    class Meta:
        model = ModelMeta
        fields = '__all__'

class ModelTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ModelType
        fields = '__all__'

class ActiveModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = ActiveModel
        fields = '__all__'


from django.db import models

class ProductManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
def sum_weights(product_materials):
    # sum up the weights
    weight = 0
    reference_unit = None
    for product_material in product_materials:
        if not reference_unit:
            reference_unit = product_material.unit
        else:
            assert reference_unit == product_material.unit, "ProductMaterial units do not match"
        weight += product_material.quantity

    return weight, reference_unit

def sum_emissions(product_materials):
    # sum up the emissions
    emission = 0
    reference_unit = None
    for product_material in product_materials:
        if not reference_unit:
            reference_unit = product_material.unit
        else:
            assert reference_unit == product_material.unit, "ProductMaterial units do not match"
        emission, emission_unit = product_material.get_emission()
        assert reference_unit == emission_unit, "ProductMaterial units do not match"
        emission += emission

    return emission, reference_unit

class Product(models.Model):
    objects = ProductManager()

    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, null=True)
    name_en = models.CharField(max_length=255, null=True)

    transport_route = models.ForeignKey("TransportRoute", on_delete=models.SET_NULL, null=True)
    product_group = models.ForeignKey("ProductGroup", on_delete=models.SET_NULL, null=True)

    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name
    
    def get_product_weight(self):
        # check if there is a product material weight
        from .product_material import ProductMaterial
        product_materials = ProductMaterial.objects.filter(product=self, component="product")
        if product_materials:
            return self.get_product_material_weight()
        
        # check if there is a product weight
        #TODO

    def get_package_weight(self):
        # check if there is a package material weight
        from .product_material import ProductMaterial
        product_materials = ProductMaterial.objects.filter(product=self, component="package")
        if product_materials:
            return self.get_package_material_weight()
        
        # check if there is a package weight
        #TODO

    def get_product_material_weight(self):
        # get all materials with component == "product"
        from .product_material import ProductMaterial
        product_materials = ProductMaterial.objects.filter(product=self, component="product")

        return sum_weights(product_materials)
    
    def get_package_material_weight(self):
        # get all materials with component == "package"
        from .product_material import ProductMaterial
        product_materials = ProductMaterial.objects.filter(product=self, component="package")

        return sum_weights(product_materials)

    def get_product_material_emission(self):
        # get all materials with component == "product"
        from .product_material import ProductMaterial
        product_materials = ProductMaterial.objects.filter(product=self, component="product")

        return sum_emissions(product_materials)

    def get_package_material_emission(self):
        # get all materials with component == "package"
        from .product_material import ProductMaterial
        product_materials = ProductMaterial.objects.filter(product=self, component="package")

        return sum_emissions(product_materials)



from django.db import models
import numpy as np
from scipy.stats import skewnorm

class BaseValueDistribution(models.Model):
    """
    Abstract base class for value distributions.
    """
    name = models.CharField(max_length=100)

    class Meta:
        abstract = True

    def generate_value(self):
        """
        Generate a value based on the distribution rules.
        Must be implemented by subclasses.
        """
        raise NotImplementedError("Subclasses must implement this method")
    
    def natural_key(self):
        return (self.name,)

class NumericValueDistributionManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class NumericValueDistribution(BaseValueDistribution):
    """
    Numeric value distribution model.
    Supports uniform, normal, and skewed normal distributions with hard limits.
    """
    objects = NumericValueDistributionManager()
    DISTRIBUTION_CHOICES = [
        ('uniform', 'Uniform'),
        ('normal', 'Normal'),
        ('skewed_normal', 'Skewed Normal'),
    ]

    distribution_type = models.CharField(max_length=20, choices=DISTRIBUTION_CHOICES)
    min_value = models.FloatField()
    max_value = models.FloatField()
    mean = models.FloatField(null=True, blank=True)
    std_dev = models.FloatField(null=True, blank=True)
    skewness = models.FloatField(null=True, blank=True)

    def generate_value(self):
        if self.distribution_type == 'uniform':
            return np.random.uniform(self.min_value, self.max_value)
        elif self.distribution_type == 'normal':
            value = np.random.normal(self.mean, self.std_dev)
            return np.clip(value, self.min_value, self.max_value)
        elif self.distribution_type == 'skewed_normal':
            value = skewnorm.rvs(a=self.skewness, loc=self.mean, scale=self.std_dev)
            return np.clip(value, self.min_value, self.max_value)
        else:
            raise ValueError("Unsupported distribution type")
        
    
class SingleCategoricalValueDistributionManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class SingleCategoricalValueDistribution(BaseValueDistribution):
    """
    Single categorical value distribution model.
    Assigns a single value based on specified probabilities.
    """
    objects = SingleCategoricalValueDistributionManager()
    categories = models.JSONField()  # { "category": "probability", ... }

    def generate_value(self):
        categories, probabilities = zip(*self.categories.items())
        return np.random.choice(categories, p=probabilities)
    

class MultipleCategoricalValueDistributionManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class MultipleCategoricalValueDistribution(BaseValueDistribution):
    """
    Multiple categorical value distribution model.
    Assigns a specific number or varying number of values based on probabilities.
    """
    objects = MultipleCategoricalValueDistributionManager()
    categories = models.JSONField()  # { "category": "probability", ... }
    min_count = models.IntegerField()
    max_count = models.IntegerField()
    count_distribution_type = models.CharField(max_length=20, choices=[('uniform', 'Uniform'), ('normal', 'Normal')])
    count_mean = models.FloatField(null=True, blank=True)
    count_std_dev = models.FloatField(null=True, blank=True)

    def generate_value(self):
        if self.count_distribution_type == 'uniform':
            count = np.random.randint(self.min_count, self.max_count + 1)
        elif self.count_distribution_type == 'normal':
            count = int(np.random.normal(self.count_mean, self.count_std_dev))
            count = np.clip(count, self.min_count, self.max_count)
        else:
            raise ValueError("Unsupported count distribution type")

        categories, probabilities = zip(*self.categories.items())
        return list(np.random.choice(categories, size=count, p=probabilities))


class DateValueDistributionManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

from datetime import date, timedelta
class DateValueDistribution(BaseValueDistribution):
    """
    Assign date values based on specified distribution.
    Expects distribution_type (uniform, normal) and mode (date, timedelta) and based on this either
    date_min, date_max, date_mean, date_std_dev or
    timedelta_days_min, timedelta_days_max, timedelta_days_mean, timedelta_days_std_dev
    """
    objects = DateValueDistributionManager()
    name = models.CharField(max_length=100)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    DISTRIBUTION_CHOICES = [
        ('uniform', 'Uniform'),
        ('normal', 'Normal'),
    ]
    MODE_CHOICES = [
        ('date', 'Date'),
        ('timedelta', 'Timedelta'),
    ]

    distribution_type = models.CharField(max_length=20, choices=DISTRIBUTION_CHOICES)
    mode = models.CharField(max_length=20, choices=MODE_CHOICES)

    # Date-related fields
    date_min = models.DateField(blank=True, null=True)
    date_max = models.DateField(blank=True, null=True)
    date_mean = models.DateField(blank=True, null=True)
    date_std_dev = models.IntegerField(blank=True, null=True)  # Standard deviation in days

    # Timedelta-related fields
    timedelta_days_min = models.IntegerField(blank=True, null=True)
    timedelta_days_max = models.IntegerField(blank=True, null=True)
    timedelta_days_mean = models.IntegerField(blank=True, null=True)
    timedelta_days_std_dev = models.IntegerField(blank=True, null=True)

    def generate_value(self):
        if self.mode == 'date':
            return self._generate_date_value()
        elif self.mode == 'timedelta':
            return self._generate_timedelta_value()
        else:
            raise ValueError("Unsupported mode")

    def _generate_date_value(self):
        #UNTESTED
        if self.distribution_type == 'uniform':
            start_date = self.date_min.toordinal()
            end_date = self.date_max.toordinal()
            random_ordinal = np.random.randint(start_date, end_date)
            return date.fromordinal(random_ordinal)
        elif self.distribution_type == 'normal':
            mean_ordinal = self.date_mean.toordinal()
            std_dev_days = self.date_std_dev
            random_ordinal = int(np.random.normal(mean_ordinal, std_dev_days))
            random_ordinal = np.clip(random_ordinal, self.date_min.toordinal(), self.date_max.toordinal())
            return date.fromordinal(random_ordinal)
        else:
            raise ValueError("Unsupported distribution type")

    def _generate_timedelta_value(self):
        if self.distribution_type == 'uniform':
            random_days = np.random.randint(self.timedelta_days_min, self.timedelta_days_max + 1)
            
            
        elif self.distribution_type == 'normal':
            random_days = int(np.random.normal(self.timedelta_days_mean, self.timedelta_days_std_dev))
            random_days = np.clip(random_days, self.timedelta_days_min, self.timedelta_days_max)
            
        else:
            raise ValueError("Unsupported distribution type")
        
        current_date = date.today()
        generated_date = current_date - timedelta(days=random_days)
        print(generated_date)
        return(generated_date)
        
# Example Usage
# Numeric distribution for age
# age_distribution = NumericValueDistribution.objects.create(
#     name='Age Distribution',
#     distribution_type='normal',
#     min_value=0,
#     max_value=100,
#     mean=50,
#     std_dev=15
# )

# # Single categorical distribution for gender
# gender_distribution = SingleCategoricalValueDistribution.objects.create(
#     name='Gender Distribution',
#     categories={'male': 0.5, 'female': 0.5}
# )

# # Multiple categorical distribution for symptoms
# symptoms_distribution = MultipleCategoricalValueDistribution.objects.create(
#     name='Symptoms Distribution',
#     categories={'fever': 0.3, 'cough': 0.4, 'fatigue': 0.2, 'nausea': 0.1},
#     min_count=1,
#     max_count=3,
#     count_distribution_type='normal',
#     count_mean=2,
#     count_std_dev=0.5
# )



from django.db import models

class EndoscopeManager(models.Manager):
    def get_by_natural_key(self, name, sn):
        return self.get(name=name, sn=sn)
    
class Endoscope(models.Model):
    objects = EndoscopeManager()
    
    name = models.CharField(max_length=255)
    sn = models.CharField(max_length=255)
    
    def natural_key(self):
        return (self.name, self.sn)
    
    def __str__(self):
        return self.name
    
    class Meta:
        ordering = ['name']
        verbose_name = 'Endoscope'
        verbose_name_plural = 'Endoscopes'

class EndoscopeTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class EndoscopeType(models.Model):
    objects = EndoscopeTypeManager()
    
    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    
    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name
    
    class Meta:
        ordering = ['name']
        verbose_name = 'Endoscope Type'
        verbose_name_plural = 'Endoscope Types'



from django.db import models
    
class CenterWaste(models.Model):
    center = models.ForeignKey("Center", on_delete=models.CASCADE)
    year = models.IntegerField()
    waste = models.ForeignKey("Waste", on_delete=models.CASCADE)
    quantity = models.FloatField()
    unit = models.ForeignKey("Unit", on_delete=models.SET_NULL, null=True)
    emission_factor = models.ForeignKey("EmissionFactor", on_delete=models.SET_NULL, null=True)

    


from endoreg_db.models import PatientExamination
from typing import Optional

def get_patient_examinations() -> PatientExamination:
    """
    Returns all PatientExamination objects from the database.
    """
    return PatientExamination.objects.all()

def get_patient_examinations_without_report_file() -> PatientExamination:
    """
    Returns all PatientExamination objects from the database without a report_file.
    """
    return PatientExamination.objects.filter(report_file__isnull=True)

def get_patient_examinations_without_video() -> PatientExamination:
    """
    Returns all PatientExamination objects from the database without a video.
    """
    return PatientExamination.objects.filter(video__isnull=True)



from django.db import models

class AglServiceManager(models.Manager):
    # Custom manager for AglService; defines name as natural key
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class AglService(models.Model):
    name = models.CharField(max_length=255)
    description = models.CharField(max_length=255)
    online = models.BooleanField(default=False)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    objects = AglServiceManager()
    devices = models.ManyToManyField("NetworkDevice", related_name="services")
    port = models.IntegerField(blank=True, null=True)
    protocol = models.CharField(max_length=255, blank=True, null=True)
    url = models.URLField(blank=True, null=True)

    def alive_log(self, user, ip_address, user_agent):
        from ..logging import AglServiceLogEntry
        AglServiceLogEntry.objects.create(
            user=user,
            ip_address=ip_address,
            user_agent=user_agent,
            message=f"Service {self.name} is alive.",
            service=self
        )

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        
    def natural_key(self):
        return (self.name,)
    



from django.db import models

YES_NO_CHOICES = (
    ('yes', 'Ja'),
    ('no', 'Nein'),
)

class TtoQuestionnaire(models.Model):

    # Patient Information
    patient_name = models.CharField(max_length=255, verbose_name="Identifikation des Patienten (Name)")
    birth_date = models.DateField(verbose_name="Geburtsdatum")

    # Base documentation
    emergency_patient = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Notfallpatient/kürzlich untersuchter Patient (Verzicht auf Team-Time-Out möglich)")
    consent_signed = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Einverständniserklärung unterschrieben (Arzt, Patient)")
    documents_present = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Alle Dokumente liegen vor (Labor, Befunde, etc.)")
    communication_possible = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Kommunikation mit Patient möglich")
    work_incapacity_certificate = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Arbeitsunfähigkeitsbescheinigung")

    # priority items
    pregnancy = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Schwangerschaft")
    asa_classification_checked = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="ASA-Klassifikation/Komorbidität geprüft")
    previous_anesthesia_complications = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Komplikationen bei bisherigen Narkosen?")
    last_meal_over_6_hours_ago = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Zeitpunkt letzte Mahlzeit > 6 Stunden")
    allergies = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Allergien, welche?")
    outpatient_accompaniment_post_sedation = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Nur bei ambulanter Vorstellung: Begleitung nach Sedierung?")
    
    # Possessions
    dental_prosthesis = models.BooleanField(default=False, verbose_name="Zahnprothese")
    glasses = models.BooleanField(default=False, verbose_name="Brille")
    implants = models.BooleanField(default=False, verbose_name="Implantate")
    hearing_aids = models.BooleanField(default=False, verbose_name="Hörgeräte")
    

    # Medical History
    anticoagulants_ass = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Antikoagulation, ASS")
    blood_pressure_medication = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Blutdruckmedikamente")
    glaucoma = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Glaukom")

    metal_implants = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Metallimplantate")
    pacemaker_icd = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Herzschrittmacher/ICD")
    
    copd = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="COPD")
    liver_cirrhosis = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Leberzirrhose")
    
    ibd = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="CED (Chronisch entzündliche Darmerkrankungen)")
    radiation = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Bestrahlung")
    surgeries = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="OP´s")

    # preflight
    team_introduction = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Teamvorstellung mit Name und Aufgabe")
    instruments_available = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Notwendige Instrumente vorhanden?")
    monitoring_medications_equipment_checked = models.CharField(max_length=3, choices=YES_NO_CHOICES, verbose_name="Monitoring, Medikamente, Equipment zum Atemwegsmanagement zur Verfügung und überprüft?")
    complete_documentation_inclusive_care_notes = models.TextField(verbose_name="Vollständige Dokumentation inklusive Hinweise für Nachsorge")
    
    # notes
    notes = models.TextField(verbose_name="Bemerkungen", default = "Keine Bemerkungen")
    
    # postflight  
    specimens_secured = models.BooleanField(default=False, verbose_name="Histologische Proben gesichert")
    patient_condition_documented = models.TextField(verbose_name="Patientenzustand dokumentiert (je nach Ausgangszustand)")

    

    def __str__(self):
        return f"Endoscopy Questionnaire for {self.patient_name} on {self.birth_date}"
    
    def get_id_attributes(self):
        _ = [self.patient_name, self.birth_date]
        return _
    
    def get_base_documentation_attributes(self):
        _ = [self.emergency_patient, self.consent_signed, self.documents_present, self.communication_possible, self.work_incapacity_certificate]
        return _
    
    def get_priority_items_attributes(self):
        _ = [self.pregnancy, self.asa_classification_checked, self.previous_anesthesia_complications, self.last_meal_over_6_hours_ago, self.allergies, self.outpatient_accompaniment_post_sedation]
        return _
    
    def get_possessions_attributes(self):
        _ = [self.dental_prosthesis, self.glasses, self.implants, self.hearing_aids]
        return _
    
    def get_medical_history_attributes(self): 
        _ = [
            self.anticoagulants_ass,
            self.blood_pressure_medication,
            self.glaucoma,
            self.metal_implants,
            self.pacemaker_icd,
            self.copd,
            self.liver_cirrhosis,
            self.ibd,
            self.radiation,
            self.surgeries,
        ]

    def get_preflight_attributes(self):
        _ = [
            self.team_introduction,
            self.instruments_available,
            self.monitoring_medications_equipment_checked,
        ]

    def get_note_attributes(self):
        _ = [self.notes]

    def get_postflight_attributes(self):
        _ = [
            self.specimens_secured,
            self.patient_condition_documented
        ]




# command expects a directory as an argument
# directory should contain a img_dicts.jsonl file, a settings.json and an "images" directory
# images should be in jpeg format
# reads the img_dicts.jsonl file and creates a new database entry for each image
# img_dicts contains a line for each image in the following format:
# {"id": 30604292, "path": "images/30604292.jpg", "labels": ["appendix", "polyp"]}

# import_settings.json example:
# {
#     "annotator": "legacy",
#     "labelset_name": "multilabel_classification",
#     "labelset_version": 0
# }

import os
import json
from tqdm import tqdm
from django.core.management.base import BaseCommand
from endoreg_db.models import LegacyImage, LabelSet, Label, ImageClassificationAnnotation
from django.core.files import File
from uuid import uuid4

class Command(BaseCommand):
    """
    Imports images from a directory into the database

    Usage:
        python manage.py import_legacy_images <directory>
    """
    help = 'Imports images from a directory into the database'

    def add_arguments(self, parser):
        parser.add_argument('directory', type=str)

    def handle(self, *args, **options):
        directory = options['directory']
        if not os.path.isdir(directory):
            raise Exception(f"Directory {directory} does not exist")

        # read settings.json
        with open(os.path.join(directory, 'import_settings.json'), 'r') as f:
            settings = json.load(f)

        # read img_dicts.jsonl
        with open(os.path.join(directory, 'img_dicts.jsonl'), 'r') as f:
            img_dicts = [json.loads(line) for line in f.readlines()]

        # get labelset
        labelset_name = settings['labelset_name']
        labelset_version = settings['labelset_version']
        try:
            labelset = LabelSet.objects.get(name=labelset_name, version=labelset_version)
        except LabelSet.DoesNotExist:
            raise Exception(f"No labelset found with the name {labelset_name} and version {labelset_version}")


        # get labels in dict to lookup by name
        labels = {label.name: label for label in labelset.labels.all()}

        # create images and image_classification_annotations
        for img_dict in tqdm(img_dicts):

            # Open the image file
            with open(os.path.join(directory, img_dict['path']), 'rb') as f:
                # Create the Django File object
                django_file = File(f)
                
                # Extract only the filename
                filename = os.path.basename(img_dict['path'])
                img_suffix = os.path.splitext(filename)[1]
                # assign uuid as new filename
                filename = str(uuid4()) + img_suffix
                

                # Create a new LegacyImage instance and save the image
                image = LegacyImage(image=filename, suffix=img_suffix)
                image.image.save(filename, django_file, save=True)

                image_annotations = []
                for label_name, label in labels.items():
                    if label_name in img_dict['labels']:
                        value = True
                    else:
                        value = False

                    image_annotations.append(ImageClassificationAnnotation(
                        legacy_image=image,
                        label=label,
                        value=value,
                        annotator=settings['annotator']
                    ))

                ImageClassificationAnnotation.objects.bulk_create(image_annotations)




import hashlib
from pathlib import Path

def get_video_hash(video_path):
    """
    Get the hash of a video file.
    """
    # Open the video file in read-binary mode:
    with open(video_path, 'rb') as f:
        # Create the hash object, passing in the video contents for hashing:
        hash_object = hashlib.sha256(f.read())
        # Get the hexadecimal representation of the hash
        video_hash = hash_object.hexdigest()
        assert len(video_hash) <= 255, "Hash length exceeds 255 characters"

    return video_hash

def get_pdf_hash(pdf_path:Path):
    """
    Get the hash of a pdf file.
    """
    pdf_hash = None

    # Open the file in binary mode and read its contents
    with open(pdf_path, 'rb') as f:
        pdf_contents = f.read()
        # Create a hash object using SHA-256 algorithm

    hash_object = hashlib.sha256(pdf_contents, usedforsecurity=False)
    # Get the hexadecimal representation of the hash
    pdf_hash = hash_object.hexdigest()
    assert len(pdf_hash) <= 255, "Hash length exceeds 255 characters"

    return pdf_hash


from django.db import models

class ImageClassificationAnnotation(models.Model):
    """
    A class representing an image classification annotation.

    Attributes:
        label (str): The label that was assigned to the image.
        confidence (float): The confidence that the label is correct.
        annotator (User): The user that created this annotation.
        date (datetime.datetime): The date and time when this annotation was created.

    """
    # Foreign keys to Frame, LegacyFrame, and LegacyImage (only one of these should be set)
    frame = models.ForeignKey("Frame", on_delete=models.CASCADE, blank=True, null=True, related_name="image_classification_annotations")
    legacy_frame = models.ForeignKey("LegacyFrame", on_delete=models.CASCADE, blank=True, null=True, related_name="image_classification_annotations")
    legacy_image = models.ForeignKey("LegacyImage", on_delete=models.CASCADE, blank=True, null=True, related_name="image_classification_annotations")

    label = models.ForeignKey("Label", on_delete=models.CASCADE, related_name="image_classification_annotations")
    value = models.BooleanField()
    annotator = models.CharField(max_length=255)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.label.name + " - " + str(self.value)
    


from django.db import models

# import endoreg_center_id from django settings
from django.conf import settings


# import File class
from django.core.files import File

# # check if endoreg_center_id is set
# if not hasattr(settings, 'ENDOREG_CENTER_ID'):
#     ENDOREG_CENTER_ID = 9999
# else:
#     ENDOREG_CENTER_ID = settings.ENDOREG_CENTER_ID

class PdfType(models.Model):
    name = models.CharField(max_length=255)

    patient_info_line = models.ForeignKey(
        "ReportReaderFlag",
        related_name="pdf_type_patient_info_line", 
        on_delete=models.CASCADE
    )
    endoscope_info_line = models.ForeignKey(
        "ReportReaderFlag",
        related_name="pdf_type_endoscopy_info_line", 
        on_delete=models.CASCADE,
    )
    examiner_info_line = models.ForeignKey(
        "ReportReaderFlag",
        related_name="pdf_type_examiner_info_line", 
        on_delete=models.CASCADE
    )
    cut_off_above_lines = models.ManyToManyField(
        "ReportReaderFlag",
        related_name="pdf_type_cut_off_above_lines", 
    )
    cut_off_below_lines = models.ManyToManyField(
        "ReportReaderFlag",
        related_name="pdf_type_cut_off_below_lines", 
    )


    def __str__(self):
        summary = f"{self.name}"
        # add lines to summary
        summary += f"\nPatient Info Line: {self.patient_info_line.value}"
        summary += f"\nEndoscope Info Line: {self.endoscope_info_line.value}"
        summary += f"\nExaminer Info Line: {self.examiner_info_line.value}"
        summary += f"\nCut Off Above Lines: {[_.value for _ in self.cut_off_above_lines.all()]}"
        summary += f"\nCut Off Below Lines: {[_.value for _ in self.cut_off_below_lines.all()]}"
    
        return summary

class PdfMeta(models.Model):
    pdf_type = models.ForeignKey(PdfType, on_delete=models.CASCADE)
    date = models.DateField()
    time = models.TimeField()
    pdf_hash = models.CharField(max_length=255, unique=True)

    def __str__(self):
        return self.pdf_hash

    @classmethod
    def create_from_file(cls, pdf_file):
        pdf_file = File(pdf_file)
        pdf_meta = cls(file=pdf_file)
        pdf_meta.save()
        return pdf_meta




from django.db import models
from .base import AbstractLogEntry

class NetworkDeviceLogEntry(AbstractLogEntry):
    """
    Model for log entries related to network devices.
    """
    device = models.ForeignKey("NetworkDevice", on_delete=models.CASCADE)
    log_type = models.ForeignKey("LogType", on_delete=models.CASCADE, null=True, blank=True)

    class Meta:
        verbose_name = "Network Device Log Entry"
        verbose_name_plural = "Network Device Log Entries"

    def __str__(self):
        return f"{self.timestamp} - {self.user} - {self.ip_address} - {self.device} - {self.message}"




from endoreg_db.utils.hashs import get_video_hash
import subprocess
from django.core.files import File
from django.core.management.base import BaseCommand
from endoreg_db.models.data_file.video import LegacyVideo
import os
from tqdm import tqdm
# import cv2

def convert_mkv_to_mp4(source_path, target_path):
    """
    Convert a .mkv file to a .mp4 file using FFmpeg.
    
    Parameters:
        mov_file_path (str): The file path of the input .MOV file.
        mp4_file_path (str): The file path where the output .mp4 file should be saved.
        
    Returns:
        None
    """
    cmd = ["ffmpeg", "-y", "-i", source_path, "-vcodec", "h264", "-acodec", "aac", target_path]
    subprocess.run(cmd)


class Command(BaseCommand):
    """
    Imports videos from a directory into the database

    Usage:
        python manage.py import_legacy_videos <directory>
    """
    help = 'Imports videos from a directory into the database'

    def add_arguments(self, parser):
        parser.add_argument('directory', type=str)

    def handle(self, *args, **options):
        directory = options['directory']
        if not os.path.isdir(directory):
            raise Exception(f"Directory {directory} does not exist")

        # iterate over all subdirectories and gather all .mkv files
        mkv_files = []
        for root, dirs, files in os.walk(directory):
            for file in files:
                if file.endswith(".mkv"):
                    mkv_files.append(os.path.join(root, file))

        # create videos
        for mkv_file in tqdm(mkv_files):
            # get the name of the video
            video_name = os.path.basename(mkv_file).replace(".mkv", "")

            # convert the .mkv file to a .mp4 file
            mp4_file = os.path.join(directory, video_name + ".mp4")
            if not os.path.isfile(mp4_file):
                convert_mkv_to_mp4(mkv_file, mp4_file)

            # create the video object
            video_hash = get_video_hash(mp4_file)

            # check if the video already exists
            if LegacyVideo.objects.filter(video_hash=video_hash).exists():
                print(f"Video with hash {video_hash} already exists. Skipping...")
                continue

            video = LegacyVideo.objects.create(video_hash=video_hash, suffix = "mp4")

            # open the .mp4 file
            with open(mp4_file, 'rb') as f:
                # create the Django File object
                django_file = File(f)
                # save the file to the video
                video.file.save(video_name + ".mp4", django_file)
            raise Exception("Stop here: NEED TO FIX OPENCV DEPENDENCY")
            video.initialize_video_specs(cv2.VideoCapture(video.file.path))



from ..raw_video import RawVideoFile
import json 

# # Starting point
# Automated tasks generate RawVideoFile objects in our db.
# Each object has state_{NAME} attributes.
# We will create functions which query the db for RawVideoFile 
# objects with specific state_{NAME} attributes.
# Then, we perform the necessary operations on the RawVideoFile and
# update the state_{NAME} attributes accordingly.

# # Step 1 - Frame Extraction
# function to query for videos scheduled for frame extraction,
# these have state_frames_required and state_frames_extracted
def get_videos_scheduled_for_frame_extraction():
    return RawVideoFile.objects.filter(
        state_frames_required=True,
        state_frames_extracted=False
    )

def extract_frames_from_video(video:RawVideoFile):
    # extract frames from video
    video.extract_frames()

    # update state_frames_extracted
    video.state_frames_extracted = True
    video.save()

    return video

def extract_frames_from_videos():
    videos = get_videos_scheduled_for_frame_extraction()
    for video in videos:
        extract_frames_from_video(video)

# # Step 2 - OCR
# function to query for videos scheduled for OCR,
# these have 
# state_ocr_required = True and state_ocr_completed = False and state_frames_extracted = True
def get_videos_scheduled_for_ocr():
    return RawVideoFile.objects.filter(
        state_ocr_required=True,
        state_ocr_completed=False,
        state_frames_extracted=True
    )

# function to set state_frames_required to True for videos 
# which are scheduled for OCR but have not had frames extracted
def videos_scheduled_for_ocr_preflight():
    videos = RawVideoFile.objects.filter(
        state_ocr_required=True,
        state_ocr_completed=False,
        state_frames_extracted=False
    )
    for video in videos:
        video.state_frames_required = True
        video.save()

def perform_ocr_on_video(video:RawVideoFile):
    # perform OCR on video
    video.update_text_metadata()

    # update state_ocr_completed
    video.state_ocr_completed = True
    video.save()

    return video

def perform_ocr_on_videos():
    videos = get_videos_scheduled_for_ocr()
    for video in videos:
        perform_ocr_on_video(video)


# # Step 3 - initial Prediction
# function to query for videos scheduled for initial prediction,
# these have
# state_initial_prediction_required = True and state_initial_prediction_completed = False and state_frames_extracted = True
def videos_scheduled_for_initial_prediction_preflight():
    videos = RawVideoFile.objects.filter(
        state_initial_prediction_required=True,
        state_initial_prediction_completed=False,
        state_frames_extracted=False
    )
    for video in videos:
        video.state_frames_required = True
        video.save()

def get_videos_scheduled_for_initial_prediction():
    return RawVideoFile.objects.filter(
        state_initial_prediction_required=True,
        state_initial_prediction_completed=False,
        state_frames_extracted=True
    )

from pathlib import Path
def get_multilabel_model(model_path:Path):
    from agl_predict_endo_frame.model_loader import MultiLabelClassificationNet
    model_path_str = model_path.resolve().as_posix()
    model = MultiLabelClassificationNet.load_from_checkpoint(model_path_str)
    model.cuda()
    model.eval()
    return model

def get_multilabel_classifier(model, verbose:bool=False):
    from agl_predict_endo_frame.predict import Classifier
    classifier = Classifier(model, verbose = verbose)
    return classifier

def get_crops(video, paths):
    endo_roi_dict = video.get_endo_roi()
    # dict with x, y, width height
    # crops is list of touples with (y_min, y_max, x_min, x_max)
    crop_tuple = (
        endo_roi_dict["y"],
        endo_roi_dict["y"] + endo_roi_dict["height"],
        endo_roi_dict["x"],
        endo_roi_dict["x"] + endo_roi_dict["width"],  
    )
    crops = [crop_tuple for _ in paths]
    return crops

# model = MultiLabelClassificationNet.load_from_checkpoint("model/colo_segmentation_RegNetX800MF_6.ckpt")
def perform_initial_prediction_on_video(
    video:RawVideoFile, model_path,
    window_size_s, min_seq_len_s
):

    model = get_multilabel_model(model_path)
    classifier = get_multilabel_classifier(model, verbose = True)

    paths = video.get_frame_paths()
    string_paths = [p.resolve().as_posix() for p in paths]
    crops = get_crops(video, string_paths)
    fps = video.get_fps()

    predictions = classifier.pipe(string_paths, crops)
    readable_predictions = [classifier.readable(p) for p in predictions]
    result_dict = classifier.post_process_predictions_serializable(
        readable_predictions,
        window_size_s = window_size_s,
        min_seq_len_s = min_seq_len_s,
        fps = fps
    )

    
    # Predictions
    _path = video.get_predictions_path()
    with open(_path, "w") as f:
        json.dump(result_dict["predictions"], f, indent = 4)

    # smooth_predictions
    _path = video.get_smooth_predictions_path()
    with open(_path, "w") as f:
        json.dump(result_dict["smooth_predictions"], f, indent = 4)

    # binary_predictions
    _path = video.get_binary_predictions_path()
    with open(_path, "w") as f:
        json.dump(result_dict["binary_predictions"], f, indent = 4)

    # Raw Sequences
    _path = video.get_raw_sequences_path()
    with open(_path, "w") as f:
        json.dump(result_dict["raw_sequences"], f, indent = 4)

    # filtered_sequences
    _path = video.get_filtered_sequences_path()
    with open(_path, "w") as f:
        json.dump(result_dict["filtered_sequences"], f, indent = 4)

    
    # update state_initial_prediction_completed
    video.state_initial_prediction_required = False
    video.state_initial_prediction_completed = True
    video.state_initial_prediction_import_required = True
    video.state_initial_prediction_import_completed = False
    video.save()

    return video

def perform_initial_prediction_on_videos(
    model_path,
    window_size_s, min_seq_len_s
):
    videos = get_videos_scheduled_for_initial_prediction()
    for video in videos:
        perform_initial_prediction_on_video(
        video,
        model_path, window_size_s, min_seq_len_s
    )

def videos_scheduled_for_prediction_import_preflight():
    videos = RawVideoFile.objects.filter(
        state_initial_prediction_completed=True,
        state_initial_prediction_import_completed=False
    )
    for video in videos:
        video.state_initial_prediction_required = True
        video.save()
    
def get_videos_scheduled_for_prediction_import():
    return RawVideoFile.objects.filter(
        state_initial_prediction_import_required=True,
        state_initial_prediction_import_completed=False,
        state_initial_prediction_completed=True
    )

def import_predictions_for_video(video:RawVideoFile):
    # import predictions for video
    pass

    # update state_prediction_import_completed
    # video.state_initial_prediction_import_required = False
    # video.state_initial_prediction_import_completed = True
    # video.save()

    return video

def import_predictions_for_videos():
    videos = get_videos_scheduled_for_prediction_import()
    for video in videos:
        import_predictions_for_video(video)

# # Step 4 - Delete Frames if not needed anymore
# function to query for videos scheduled for frame deletion,
# first we need to set state_frames_required = False for videos with:
# state_ocr_required = False and state_ocr_completed = True and
# state_initial_prediction_required = False and state_initial_prediction_completed = True
def delete_frames_preflight():
    videos = RawVideoFile.objects.filter(
        state_ocr_required=False,
        state_ocr_completed=True,
        state_initial_prediction_required=False,
        state_initial_prediction_completed=True
    )
    for video in videos:
        video.state_frames_required = False
        video.save()

# function to query for videos scheduled for frame deletion,
# frames should be deleted if state_frames_required = False
def get_videos_scheduled_for_frame_deletion():
    return RawVideoFile.objects.filter(
        state_frames_required=False
    )

def delete_frames_for_video(video:RawVideoFile):
    # delete frames for video

    # update state_frames_deleted
    video.state_frames_extracted = False
    video.save()

    return video

def delete_frames():
    videos = get_videos_scheduled_for_frame_deletion()
    for video in videos:
        delete_frames_for_video(video)



from pathlib import Path

DATA_DIR = Path(__file__).parent

INFORMATION_DATA_DIR = DATA_DIR / "information_source"
UNIT_DATA_DIR = DATA_DIR / "unit"
CENTER_DATA_DIR = DATA_DIR / "center"
EXAMINATION_DATA_DIR = DATA_DIR / "examination"
LABEL_DATA_DIR = DATA_DIR / "label"
MODEL_TYPE_DATA_DIR = DATA_DIR / "model_type"
ENDOSCOPY_PROCESSOR_DATA_DIR = DATA_DIR / "endoscopy_processor"
ENDOSCOPE_TYPE_DATA_DIR = DATA_DIR / "endoscope_type"
ACTIVE_MODEL_DATA_DIR = DATA_DIR / "active_model"
PROFESSION_DATA_DIR = DATA_DIR / "profession"
PDF_TYPE_DATA_DIR = DATA_DIR / "pdf_type"
REPORT_READER_FLAG_DATA_DIR = DATA_DIR / "report_reader_flag"
EVENT_DATA_DIR = DATA_DIR / "event"
GENDER_DATA_DIR = DATA_DIR / "gender"

# Green Endoscopy Dashboard Stuffs
CENTER_RESOURCE_DATA_DIR = DATA_DIR / "center_resource"
CENTER_WASTE_DATA_DIR = DATA_DIR / "center_waste"
EMISSION_FACTOR_DATA_DIR = DATA_DIR / "emission_factor"
MATERIAL_DATA_DIR = DATA_DIR / "material"
PRODUCT_DATA_DIR = DATA_DIR / "product"
PRODUCT_GROUP_DATA_DIR = DATA_DIR / "product_group"
PRODUCT_MATERIAL_DATA_DIR = DATA_DIR / "product_material"
PRODUCT_WEIGHT_DATA_DIR = DATA_DIR / "product_weight"
REFERENCE_PRODUCT_DATA_DIR = DATA_DIR / "reference_product"
RESOURCE_DATA_DIR = DATA_DIR / "resource"
WASTE_DATA_DIR = DATA_DIR / "waste"
TRANSPORT_ROUTE_DATA_DIR = DATA_DIR / "transport_route"
DISEASE_CLASSIFICATION_DATA_DIR = DATA_DIR / "disease_classification"
DISEASE_CLASSIFICATION_CHOICE_DATA_DIR = DATA_DIR / "disease_classification_choice"
DISEASE_DATA_DIR = DATA_DIR / "disease"
MEDICATION_DATA_DIR = DATA_DIR / "medication"
MEDICATION_INDICATION_DATA_DIR = DATA_DIR / "medication_indication"
MEDICATION_INDICATION_TYPE_DATA_DIR = DATA_DIR / "medication_indication_type"

DISTRIBUTION_BASE_DIR = DATA_DIR / "distribution"
DISTRIBUTION_SINGLE_CATEGORICAL_DATA_DIR = DISTRIBUTION_BASE_DIR / "single_categorical"
DISTRIBUTION_NUMERIC_DATA_DIR = DISTRIBUTION_BASE_DIR / "numeric"
DISTRIBUTION_MULTIPLE_CATEGORICAL_DATA_DIR = DISTRIBUTION_BASE_DIR / "multiple_categorical"
DISTRIBUTION_DATE_DATA_DIR = DISTRIBUTION_BASE_DIR / "date"

# G-Play Stuffs
CASE_TEMPLATE_BASE_DIR = DATA_DIR / "case_template"
CASE_TEMPLATE_DATA_DIR = CASE_TEMPLATE_BASE_DIR / "template"
CASE_TEMPLATE_TYPE_DATA_DIR = CASE_TEMPLATE_BASE_DIR / "template_type"
CASE_TEMPLATE_RULE_DATA_DIR = CASE_TEMPLATE_BASE_DIR / "rule"
CASE_TEMPLATE_RULE_TYPE_DATA_DIR = CASE_TEMPLATE_BASE_DIR / "rule_type"
CASE_TEMPLATE_RULE_VALUE_DATA_DIR = CASE_TEMPLATE_BASE_DIR / "rule_value"
CASE_TEMPLATE_RULE_VALUE_TYPE_DATA_DIR = CASE_TEMPLATE_BASE_DIR / "rule_value_type"

# NETWORK
NETWORK_DEVICE_DATA_DIR = DATA_DIR / "network_device"
NETWORK_DEVICE_TYPE_DATA_DIR = DATA_DIR / "network_device_type"
AGL_SERVICE_DATA_DIR = DATA_DIR / "agl_service"

# LOGS
LOG_TYPE_DATA_DIR = DATA_DIR / "log_type"

# To Do
LAB_VALUE_DATA_DIR = DATA_DIR / "lab_value"
PATIENT_LAB_SAMPLE_TYPE_DATA_DIR = DATA_DIR / "patient_lab_sample_type"
MEDICATION_SCHEDULE_DATA_DIR = DATA_DIR / "medication_schedule"
MEDICATION_INTAKE_TIME_DATA_DIR = DATA_DIR / "medication_intake_time"


RULE_TYPE_DATA_DIR = DATA_DIR / "rule_type"
RULE_ATTRIBUTE_DTYPE_DATA_DIR = DATA_DIR / "rule_attribute_dtype"
RULESET_DATA_DIR = DATA_DIR / "ruleset"


import pytesseract
from PIL import Image, ImageOps, ImageFilter
import os
from collections import Counter
from tempfile import TemporaryDirectory
import re
from datetime import datetime
from typing import Dict, List
from icecream import ic
import numpy as np
from endoreg_db.utils.cropping import crop_and_insert



N_FRAMES_MEAN_OCR = 2

# Helper function to process date strings
def process_date_text(date_text):
    """
    Processes a string of text that represents a date and returns a datetime.date object.

    Args:
        date_text (str): A string of text that represents a date.

    Returns:
        datetime.date: A datetime.date object representing the parsed date, or None if the text cannot be parsed.
    """
    try:
        # Remove any non-digit characters
        date_text_clean = re.sub(r'\D', '', date_text)
        # Reformat to 'ddmmyyyy' if necessary
        if len(date_text_clean) == 8:
            return datetime.strptime(date_text_clean, "%d%m%Y").date()
        elif len(date_text_clean) == 14:
            return datetime.strptime(date_text_clean, "%d%m%Y%H%M%S").date()
    except ValueError:
        # Return None if the text cannot be parsed into a date
        # set date to 1/1/1900
        return datetime.strptime("01011900", "%d%m%Y").date()

# Helper function to process patient names
def process_name_text(name_text):
    """
    Remove all numbers, punctuation, and whitespace from a string of text and return the result.
    """
    name = re.sub(r'[0-9!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~\s]+', '', name_text).strip()
    # capitalize first letter of each word
    name = ' '.join([word.capitalize() for word in name.split()])
    return name


# Helper function to process endoscope type text
def process_general_text(endoscope_text):
    """
    This function takes in a string of text from an endoscope and returns a cleaned version of the text.
    """
    return ' '.join(endoscope_text.split())

def roi_values_valid(roi):
    """
    Check if all values in an ROI dictionary are valid (>=0).
    """
    return all([value >= 0 for value in roi.values()])

# Function to extract text from ROIs
def extract_text_from_rois(image_path, processor):
    """
    Extracts text from regions of interest (ROIs) in an image using OCR.

    Args:
        image_path (str): The path to the image file.
        processor (EndoscopyProcessor): An instance of the EndoscopyProcessor class.

    Returns:
        dict: A dictionary containing the extracted text for each ROI.
    """
    # Read the image using Pillow
    image = Image.open(image_path)
    image_dimensions = image.size # (width, height)

    ####### Adjust Image #######
    # Convert to grayscale
    gray = image.convert('L')

    # Invert colors for white text on black background
    inverted = ImageOps.invert(gray)

    # Initialize the dictionary to hold the extracted text
    extracted_texts = {}

    # Define your ROIs and their corresponding post-processing functions in tuples
    rois_with_postprocessing = [
        ('examination_date', processor.get_roi_examination_date, process_date_text),
        ("patient_first_name", processor.get_roi_patient_first_name, process_name_text),
        ('patient_last_name', processor.get_roi_patient_last_name, process_name_text),
        ('patient_dob', processor.get_roi_patient_dob, process_date_text),
        ('endoscope_type', processor.get_roi_endoscope_type, process_general_text),
        ('endoscope_sn', processor.get_roi_endoscopy_sn, process_general_text),
    ]

    # Extract and post-process text for each ROI
    for roi_name, roi_function, post_process in rois_with_postprocessing:
        # Get the ROI dictionary
        roi = roi_function()
        
        # Check if the ROI has values
        
        if roi_values_valid(roi):
            x, y, w, h = roi['x'], roi['y'], roi['width'], roi['height']
            
            # Get white image with original shape and just the roi remaining
            roi_image = crop_and_insert(inverted, x,y,h,w)

            # OCR configuration: Recognize white text on black background without corrections
            config = '--psm 10 -c tessedit_char_whitelist=0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-üöäÜÖÄß'

            # Use pytesseract to do OCR on the preprocessed ROI
            text = pytesseract.image_to_string(roi_image, config=config).strip()

            # Post-process extracted text
            processed_text = post_process(text)
            
            extracted_texts[roi_name] = processed_text

        else:
            ic(roi_name)
            ic(roi)
            ic("No values for this ROI")

    return extracted_texts

def get_most_frequent_values(rois_texts: Dict[str, List[str]]) -> Dict[str, str]:
    """
    Given a dictionary of ROIs and their corresponding texts, returns a dictionary of the most frequent text for each ROI.

    Args:
        rois_texts: A dictionary where the keys are the names of the ROIs and the values are lists of texts.

    Returns:
        A dictionary where the keys are the names of the ROIs and the values are the most frequent text for each ROI.
    """
    most_frequent = {}
    for key in rois_texts.keys():
        counter = Counter([text for text in rois_texts[key] if text])
        ic(key)
        ic(counter)
        most_frequent[key], _ = counter.most_common(1)[0] if counter else (None, None)
    return most_frequent

def process_video(video_path, processor):
    """
    Processes a video file by extracting text from regions of interest (ROIs) in each frame.

    Args:
        video_path (str): The path to the video file to process.
        processor (OCRProcessor): An instance of the OCRProcessor class that defines the ROIs to extract text from.

    Returns:
        dict: A dictionary containing the most frequent text values extracted from each ROI.
    """
    # Create a temporary directory to store frames
    with TemporaryDirectory() as temp_dir:
        ic(temp_dir)
        # Capture the video
        video = cv2.VideoCapture(video_path)
        success, frame_number = True, 0
        rois_texts = {roi_name: [] for roi_name in processor.get_rois().keys()}
        frames_for_mean_extraction = 0

        while success:
            success, frame = video.read()

            # Check if this is the 200th frame
            if frame_number % 1000 == 0 and success:
                frame_path = os.path.join(temp_dir, f"frame_{frame_number}.jpg")
                cv2.imwrite(frame_path, frame)  # Save the frame as a JPEG file
                # cv2.imwrite(f"_tmp/frame_{frame_number}.jpg", frame)
                
                # Extract text from ROIs
                extracted_texts = extract_text_from_rois(frame_path, processor)
                ic(extracted_texts)
                
                # Store the extracted text from each ROI
                for key, text in extracted_texts.items():
                    rois_texts[key].append(text)
                frames_for_mean_extraction += 1

            frame_number += 1

            if frames_for_mean_extraction >= N_FRAMES_MEAN_OCR: break

        # Release the video capture object
        video.release()

        # Get the most frequent values for each ROI
        return get_most_frequent_values(rois_texts)

